/*
* This file is part of Firestorm NIDS
* Copyright (c) 2001 Gianni Tedesco
*
*    Firestorm is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*    Firestorm is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with Firestorm; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <conf.h>
#include <s_util.h>
#include <sensor.h>
#include <cleanup.h>
#include <loader.h>

conf_kw_t keywords[]={
	{"load_plugins", 	confh_load_plugins},
	{"capdev_add", 		confh_capdev_add},
	{"load_ruleset", 	confh_load_ruleset},
	{"firestorm_root",	confh_firestorm_root},
	{"chroot",		confh_chroot},
	{"effective_user", 	confh_euid},
	{"effective_group",	confh_egid},
	{"plugin_require", 	confh_require},
	{"global",		confh_global},
	{NULL, NULL}
};

kw_ret confh_require(int argc, char **argv)
{
	char *plugin_name;
	uint min,maj;
	req_t match;
	char *tmp;
	p_require r;
	
	if ( argc != 4 && argc != 2) return KW_ERROR;
	plugin_name = argv[1];
	
	if ( argc == 2 ) { 
		match = MATCH_NONE;
	}else{
		if ( !strcmp(argv[2], ">")) {
			match = MATCH_GT;
		}else if ( !strcmp(argv[2], "<")) {
			match = MATCH_LT;
		}else if ( !strcmp(argv[2], ">=")) {
			match = MATCH_GTE;
		}else if ( !strcmp(argv[2], "<=")) {
			match = MATCH_LTE;
		}else if ( !strcmp(argv[2], "=")) {
			match = MATCH_EQ;
		}else if ( !strcmp(argv[2], "==")) {
			match = MATCH_EQ;
		}else if ( !strcmp(argv[2], "!=")) {
			match = MATCH_NEQ;
		}else{
			return KW_ERROR;
		}
	}

	if ( !(tmp=strchr(argv[3], '.')) ) return KW_ERROR;
	*tmp='\0';tmp++;

	min=atoi(tmp);
	maj=atoi(argv[3]);

	if ( !(r=globals.malloc(sizeof(require_t))) ) return KW_ERROR;
	if ( !(r->pname=globals.fs_strdup(plugin_name)) ) {
		globals.free(r);
		return KW_ERROR;
	}
	r->minor=min;
	r->major=maj;
	r->match=match;
	
	globals.requires=list_append(r, globals.requires);
	
	return KW_OK;
}

kw_ret confh_global(int argc, char **argv)
{
	char *k, *v;
	p_kv t;
	
	if ( argc != 3 ) return KW_ERROR;
	if ( !(t=globals.malloc(sizeof(kv_t))) ) return KW_ERROR;
	if ( !(k=globals.fs_strdup(argv[1])) ) {
		globals.free(t);
		return KW_ERROR;
	}
	if ( !(v=globals.fs_strdup(argv[2])) ) {
		globals.free(t);
		globals.free(k);
		return KW_ERROR;
	}
	
	t->key=k;
	t->val=v;
	globals.plugin_vars=list_append(t, globals.plugin_vars);
	
	return KW_OK;
}

kw_ret confh_euid(int argc, char **argv)
{
	static int done=0;

	if ( done ) return KW_MULTIPLE;
	if ( argc != 2 ) return KW_ERROR;

	globals.uid=atoi(argv[1]);
	return KW_OK;
}

kw_ret confh_egid(int argc, char **argv)
{
	static int done=0;

	if ( done ) return KW_MULTIPLE;
	if ( argc != 2 ) return KW_ERROR;

	globals.gid=atoi(argv[1]);
	return KW_OK;
}


kw_ret confh_chroot(int argc, char **argv)
{
	static int done=0;

	if ( done ) return KW_MULTIPLE;
	if ( argc != 2 ) return KW_ERROR;

	if ( !strcasecmp("yes", argv[1]) ) {
		globals.chroot=1;
	}else{
		globals.chroot=0;
	}
	
	done=1;
	return KW_OK;
}

kw_ret confh_firestorm_root(int argc, char **argv)
{
	static int done=0;

	if ( done ) return KW_MULTIPLE;
	if ( argc != 2 ) return KW_ERROR;

	if ( !(globals.root_dir=globals.fs_strdup(argv[1])) ) return KW_ERROR;
	
	done=1;
	return KW_OK;
}

kw_ret confh_load_ruleset(int argc, char **argv) {
	char *k, *v;
	p_kv t;
	
	if ( argc != 3 ) return KW_ERROR;
	if ( !(t=globals.malloc(sizeof(kv_t))) ) return KW_ERROR;
	if ( !(k=globals.fs_strdup(argv[1])) ) {
		globals.free(t);
		return KW_ERROR;
	}
	if ( !(v=globals.fs_strdup(argv[2])) ) {
		globals.free(t);
		globals.free(k);
		return KW_ERROR;
	}
	
	t->key=k;
	t->val=v;
	globals.rulesets=list_append(t, globals.rulesets);
	
	return KW_OK;
}

kw_ret confh_capdev_add(int argc, char **argv)
{
	p_c_capdev tmp;
	
	if ( argc!=3 && argc!=4 ) return KW_ERROR;
	if ( !(tmp=globals.malloc(sizeof(c_capdev_t))) ) return KW_ERROR;
	
	tmp->type=globals.fs_strdup(argv[1]);
	tmp->source=globals.fs_strdup(argv[2]);
	tmp->filter = (argc==4) ? globals.fs_strdup(argv[3]) : NULL;
	
	if ( (tmp->type==NULL) ||
		(tmp->source==NULL) ||
		(argc==4 && tmp->filter==NULL) ) {
		if (tmp->type) globals.free(tmp->type);
		if (tmp->source) globals.free(tmp->source);
		if (tmp->filter) globals.free(tmp->filter);
		return KW_ERROR;
	}

	globals.capdevs=list_append(tmp, globals.capdevs);
	
	return KW_OK;
}

kw_ret confh_load_plugins(int argc, char **argv)
{
	p_config_arg pdir;

	/* Only want 1 argument */
	if ( argc != 2 ) return KW_ERROR;
	
	if ( !(pdir=globals.malloc(sizeof(config_arg_t))) ) return KW_ERROR;
	
	if ( !(pdir->arg=globals.fs_strdup(argv[1])) ){
		globals.free(pdir);
		return KW_ERROR;
	}
	
	globals.plugin_dirs=list_append(pdir, globals.plugin_dirs);
	
	return KW_OK;
}

void conf_cleanup_args(p_config_arg p)
{
	p_config_arg foo,bar;
	if (p==NULL) return;

	for(foo=p; foo; ){
		bar=foo;
		foo=(p_config_arg)bar->l.next;
		globals.free(bar);
		bar=NULL;
	}
}

proc_keyword_handler conf_get_handler(const char *kw)
{
	int i=0;
	proc_keyword_handler ret;
	
	if ( !kw ) return NULL;

	while ( (ret=keywords[i].proc) ) {
		if ( !strcasecmp(keywords[i].keyword, kw) ) {
			return ret;
		}
		i++;
	}
	
	return NULL;
}

int conf_parse_line(char *line)
{
	int ret;
	char *kw;
	char *tmp;
	p_config_arg args=NULL, foo;
	proc_keyword_handler kwh;
	
	int argc=1, i=1;
	char **argv;
	
	if ( !(kw=strtok(line, " \t\r\n")) ) return -1;

	if ( !(kwh=conf_get_handler(kw)) ) {
		print_out(PRINT_WARNING "Unrecognised keyword '%s'" CRLF, kw);
		return -2;
	}
	
	while ( (tmp=strtok(NULL, " \t\r\n")) ) {
		if ( !(foo=globals.malloc(sizeof(config_arg_t))) ) {
			conf_cleanup_args(args);
			return -1;
		}
		foo->arg=tmp;
		args=list_append(foo, args);
		argc++;
	}
	
	if ( !(argv=globals.malloc(argc * sizeof(char *))) ) {
		conf_cleanup_args(args);
		return -1;
	}
	
	argv[0]=kw;
	for ( foo=args; foo; foo=(p_config_arg)foo->l.next ) {
		argv[i]=foo->arg;
		i++;
	}
	
	ret = kwh(argc, argv);
	conf_cleanup_args(args);
	globals.free(argv);
	return ret;
}

int conf_load_config(char *fn)
{
	FILE *f;
	int line=0;
	char buf[1024], *p;
	
	if ( !fn ) return 0;
	
	if ( (f=fopen(fn, "r")) == NULL ) {
		print_out(PRINT_ERROR "Can't open config file %s" CRLF, fn);
		return 0;
	}
	
	for ( p=(char *)&buf; fgets(p, 1024, f); line++ ) {
		if ( !IS_CONFIG(buf) ) continue;

		switch(conf_parse_line(p)) {
			case KW_OK:
			case -2:
				break;
				
			case -1:
				print_out(PRINT_WARNING
					"Internal error in config file (%s) "
					"on line %i" CRLF, fn, line);
				return 0;

			case KW_ERROR:
				print_out(PRINT_WARNING
					"Parse error in config file (%s) "
					"on line %i" CRLF, fn, line);
				return 0;

			case KW_MULTIPLE:
				print_out(PRINT_WARNING
					"Problem in config file (%s) on line "
					"%i. This directive can only be "
					"used once." CRLF, fn, line);
				return 0;
		}
	}
	
	return 1;
}

void conf_cleanup_temp(void)
{
	if ( globals.plugin_dirs ) {
		cleanup_arg(0, globals.plugin_dirs);
		globals.plugin_dirs=NULL;
	}

	if ( globals.capdevs) {
		cleanup_c_capdev(0, globals.capdevs);
		globals.capdevs=NULL;
	}
	
	if ( globals.rulesets ) {
		cleanup_kv(0, globals.rulesets);
		globals.rulesets=NULL;
	}

	if ( globals.root_dir ) {
		memset(globals.root_dir, 0, strlen(globals.root_dir));
		globals.free(globals.root_dir);
		globals.root_dir=NULL;
	}
}

extern p_plugin_list ld_list;
int do_depend(p_require r)
{
	p_plugin_list l;

	for(l=ld_list; l; l=(p_plugin_list)l->l.next) {
		if ( !strcmp(l->name, r->pname)  ) {
			switch ( r->match ) {
				case MATCH_NONE:
					return 1;
				case MATCH_GT:
					return ( (l->ver_major>r->major) ||
						(l->ver_major>=r->major &&
						l->ver_minor>r->minor) )
						? 1 : 0;
				case MATCH_LT:
					return ( (l->ver_major<r->major) ||
						(l->ver_major<=r->major &&
						l->ver_minor<r->minor) )
						? 1 : 0;
				case MATCH_GTE:
					return ( (l->ver_major>=r->major) ||
						(l->ver_major>=r->major &&
						l->ver_minor>=r->minor) )
						? 1 : 0;
				case MATCH_LTE:
					return ( (l->ver_major<=r->major) ||
						(l->ver_major<=r->major &&
						l->ver_minor<=r->minor) )
						? 1 : 0;
				case MATCH_EQ:
					return ((l->ver_major==r->major &&
						l->ver_minor==r->minor) )
						? 1 : 0;
				case MATCH_NEQ:
					return ((l->ver_major!=r->major ||
						l->ver_minor!=r->minor) )
						? 1 : 0;
				default:
					print_out(PRINT_ERROR "Stupid plugin_require match" CRLF);
					cleanup(EXIT_ERR);
					break;
			}
		}
	}
	return 0;
}

int conf_go(void)
{
	p_require f,b;
	p_config_arg tmp=NULL;
	p_c_capdev cd=NULL;
	p_kv kv=NULL;
	
	cleanup_add_handler(cleanup_kv, (void *)globals.plugin_vars);
	
	/* Chdir to firestorm_root first */
	if ( globals.root_dir ) {
		if ((chdir(globals.root_dir) != 0)) {
			print_out(PRINT_ERROR "Could't change to "
				"firstorm root dir (%s)" CRLF,
			      	globals.root_dir);
			return 0;
		}
	}

	/* Load plugins */
	for ( tmp=globals.plugin_dirs; tmp; tmp=(p_config_arg)tmp->l.next ) {
		if ( !loader_load_dir(tmp->arg, 0) ) {
			print_out(PRINT_ERROR "EEK! Couldn't "
				"load plugin dir (%s)" CRLF, tmp->arg);
			return 0;
		}
	}
	
	/* Do the plugin dependencies */
	for ( f=(p_require)globals.requires; f; ) {
		b=f;
		f=(p_require)b->l.next;
		
		if ( !do_depend(b) ) {
			print_out(PRINT_ERROR "Plugin version requirement not met" CRLF);
			cleanup(EXIT_ERR);
		}
		
		if ( b->pname) globals.free(b->pname);
		globals.free(b);
		b=NULL;
	}
	
	/* Add and initialise all capdevs */
	for ( cd=globals.capdevs; cd; cd=(p_c_capdev)cd->l.next ) {
		if ( !capdev_add(cd->type, cd->source,
			cd->filter, capture_proc) ) {
			return 0;
		}
	}capdev_init_all();

	/* Load rulesets */
	for (kv=globals.rulesets; kv; kv=(p_kv)kv->l.next ) {
		if ( !rules_loadfile(kv->key, kv->val) ) {
			return 0;
		}
	}

	/* Chroot if necessary */
	if ( globals.chroot ) {
		if ( geteuid() ) {
			print_raw(PRINT_WARNING 
				"Cannot chroot, I'm not root." CRLF);
		}else{
			print_out("Chrooting to %s" CRLF, globals.root_dir);
			if ( chroot(".") == -1 ) {
				print_raw(PRINT_ERROR "EEK! Couldn't "
					"chroot" CRLF);
				return 0;
			}
		}
	}
	
	/* Drop UID and GID */
	if ( globals.gid ) {
		if ( geteuid() ) {
			print_raw(PRINT_WARNING 
				"Cannot change GID, I'm not root." CRLF);
		}else{
			print_out("Dropping GID to %i" CRLF, globals.gid);
			if ( setgroups(1, &globals.gid) == -1 ||
				setgid(globals.gid) == -1) {
				print_raw(PRINT_ERROR "Unexpected error "
					"while changing GID" CRLF);
				return 0;
			}
		}
	}

	if ( globals.uid ) {
		if ( geteuid() ) {
			print_raw(PRINT_WARNING
				"Cannot change UID, I'm not root." CRLF);
		}else{
			print_out("Dropping UID to %i" CRLF, globals.uid);
			if ( setuid(globals.uid) == -1 ) {
				print_raw(PRINT_ERROR "Unexpected error "
					"while changing UID" CRLF);
				return 0;
			}
		}
	}
	
	return 1;
}
